/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

#include "threadP.h"




/************************************************************************
 * reschedule.c: libthread rescheduler.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include "threadP.h"





/************************************************************************
 * void thr_reschedule(void)
 *
 * enter the thread library scheduler. save the context of the current
 * thread and start the next ready thread. the current thread is not
 * "saved" anywhere as such---the calling functions should place the
 * current thread in the proper queue before calling the scheduler.
 *
 * `thr_reschedule()' itself needs to be protected from scheduling
 * signal interrupts since it is also called from `Tcb::dummy_start()'
 * which is conceptually a user-level function.
************************************************************************/

void
thr_reschedule(void)
{
    SAVE_INLIB();

    assert(thr_curr);
    thr_check_stack();

try_to_schedule:
    if (thr_all_waitset.Count() > 0) {
        thr_service_io();
    }

    DO_TRACE_1(("thr_reschedule(): t%u leaving", thr_curr->id.tid));

#if defined(i386_unknown_nt4_0)
	volatile Thread old_thr_curr = thr_curr;
#endif // defined(i386_unknown_nt4_0)

    if (THR_SAVE_CONTEXT(thr_curr->exec.state) == 0) {
        thr_curr = (Thread) DEQUEUE(thr_ready_q);
        if (!thr_curr) {
            thr_report_error("*** no ready thread---deadlock? ***\n");
            abort();
        }
        assert(thr_curr);
        assert(thr_tab->IsThread(thr_curr->id.tid));
        THR_RESTORE_CONTEXT(thr_curr->exec.state);
    }

#if defined(i386_unknown_nt4_0)
	// In our fiber-based Windows implementation, when a thread
	// executes thr_exit() or returns from its thread function it
	// calls thr_cleanup() and then thr_reschedule().  thr_cleanup()
	// can't release the underlying fiber because a fiber that
	// calls DeleteFiber on itself will cause the entire Win32 thread 
	// to exit.  Therefore, we have to check whether the last thread
	// had exited so that we can delete its fiber for it.
	if( old_thr_curr->flags.exited )
	{
		// release its fiber
		DeleteFiber( old_thr_curr->exec.state );
		old_thr_curr->exec.state = NULL;
	}
#endif // defined(i386_unknown_nt4_0)


    if (thr_curr->flags.killed) {
        DO_TRACE_1(("thr_reschedule(): t%u killed, exiting", thr_curr->id.tid));
        thr_cleanup(0);
        goto try_to_schedule;
    }
    if (thr_curr->flags.suspended) {
        DO_TRACE_1(("thr_reschedule(): t%u suspended", thr_curr->id.tid));
        goto try_to_schedule;
    }

    DO_TRACE_1(("thr_reschedule(): t%u resuming", thr_curr->id.tid));

    RESTORE_INLIB();
    return;
}
