/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * thr.c: implementation of the Tcb class
************************************************************************/




/************************************************************************
 * header files.
************************************************************************/

#include "threadP.h"
#include <sys/types.h>

#if !defined(i386_unknown_nt4_0)
#include <sys/resource.h>
#endif /* !defined(i386_unknown_nt4_0) */





/************************************************************************
 * globals.
************************************************************************/

Tcb::Save_Func	Tcb::save_context = NULL;
Tcb::Rest_Func	Tcb::restore_context = NULL;

Tcb::Stack		Tcb::stack_pool[THR_TAB_SIZE];
unsigned		Tcb::next_stack = 0;







/************************************************************************
 * Thread Tcb::Alloc()
 *
 * Allocates a thread control block.
************************************************************************/


Thread
Tcb::Alloc( thread_t ptid,
			    Tcb::Func func,
				void* args,
				unsigned flags,
				void* stack,
				unsigned stack_size,
				unsigned is_main )
{
	// allocate storage for the thread
	thread_t tid = thr_tab->Alloc( item_t_thread );

	// construct the thread
	Thread pt = thr_tab->GetThread( tid );
	pt->Init( tid,
				ptid,
				func, args,
				flags,
				stack, stack_size,
				is_main );

	return pt;
}



void
Tcb::Init( thread_t tid,
			thread_t ptid,
			Tcb::Func func,
			void* args,
			unsigned thr_flags,
			void* stack,
			unsigned stack_size,
			unsigned is_main )
{
	unsigned i;


	id.tid       = tid;
    id.ptid         = (is_main ? tid : ptid);
    id.errno_local  = 0;

    flags.is_main   = is_main;
    flags.started   = 0;
    flags.exited    = 0;
    flags.killed    = 0;
    flags.suspended = thr_flags & THR_SUSPENDED;
    flags.blocked   = 0;
    flags.polling   = 0;
    flags.waiting   = 0;
    flags.reaped    = 0;

#if !defined(i386_unknown_nt4_0)
    exec.stack      = stack;
    exec.st_bot     = 0;
    exec.st_top     = 0;
    exec.stack_size = stack_size;
    if (!is_main && ((!exec.stack) || (!exec.stack_size))) {
        exec.stack      = GetStack( &(exec.stack_size) );
        flags.own_st    = 1;
    }
    else {
        flags.own_st = 0;
    }
	if( !is_main )
	{
		thr_adjust_stack(this);
	}
#else // !defined(i386_unknown_nt4_0)

	// In our fiber-based thread library implementation, the
	// fiber functions manage the stack.
	exec.stack		= NULL;
	exec.st_bot		= NULL;
	exec.st_top		= NULL;
	exec.stack_size	= stack_size;
	flags.own_st	= 0;
#endif // !defined(i386_unknown_nt4_0)

    call.func       = func;
    call.args       = args;
    call.result     = 0;

    wait.waiter     = 0;
    wait.departed   = 0;
    wait.status     = 0;

    for (i = 0; i < THR_KEY_TABLE_SIZE; i++) {
        key.key_data[i] = 0;
    }

    msg.mq          = 0;
	msg.tidp        = 0;
	msg.tagp        = 0;
    msg.buf         = 0;
    msg.nitemsp     = 0;

    fd.last_file    = THR_TID_UNSPEC;
	fd.last_socket  = THR_TID_UNSPEC;

    sig.nsigs       = 0;

    sprintf(id.name, "t%u", tid);
    /* NEWQ(msg.mq); */
	fd.waitset.Clear();
}






/************************************************************************
 * void dummy_start(void)
 *
 * a dummy start function for threads. this function is invoked at
 * the start of all new threads and all it does is to call the real
 * thread start function with the proper arguments.
 *
 * the current thread should not have been already started.
 *
 * this dummy function is returned-into at the completion of those
 * threads that do not use thr_exit() to exit. the dummy calls thr_exit()
 * anyway.
************************************************************************/

void
Tcb::dummy_start(void)
{
    void*  result;
    Thread t;


    assert(thr_curr);
    assert(thr_init_done);

    t = thr_curr;
    assert(!t->flags.started);

    DO_TRACE_1(("Tcb::dummy_start(): t%u, function %p starting",
        t->id.tid, t->call.func));

    t->flags.started = 1;

    DUMMY_ENTER();
    result = t->call.func(t->call.args);
    DUMMY_LEAVE();

    DO_TRACE_1(("Tcb::dummy_start(): t%u, simply returned, exiting",
        t->id.tid));

    thr_cleanup(result);
    thr_reschedule();

    /* NOTREACHED */
	thr_report_error("*** thread returned after exit ***\n");
    abort();
    return;
}






void*
Tcb::GetStack( unsigned* stack_sizep )
{
	unsigned n = next_stack;


	if( stack_pool[n].used )
	{
		n = CINCR(n, THR_TAB_SIZE);

		while( n != next_stack )
		{
			if( !(stack_pool[n].used) )
			{
				// we found an unused stack
				break;
			}
			n = CINCR(n, THR_TAB_SIZE);
		}
	}

	// we had better have found a stack - 
	// otherwise we have an inconsistency in our thread table allocation
	assert( !(stack_pool[n].used) );

	// start looking past the current stack next time we're called
	next_stack = CINCR(n, THR_TAB_SIZE);

	// return the stack
	stack_pool[n].used = true;
	*stack_sizep = DEFAULT_THR_STACK_SIZE;
	return (void*)(stack_pool[n].mem);
}



void
Tcb::Bind( thread_t tid )
{
	assert( thr_curr );

	// Add the Item to local and global waitset
	fd.waitset.Add( tid );

	// ensure our "last" indicators are valid
	if( thr_tab->IsFile( tid ) )
	{
		if( fd.last_file == THR_TID_UNSPEC )
		{
			fd.last_file = tid;
		}
	}
	else if( thr_tab->IsSocket( tid ) )
	{
		if( fd.last_socket == THR_TID_UNSPEC )
		{
			fd.last_socket = tid;
		}
	}
}


/************************************************************************
 * void Tcb::Unbind( thread_t tid )
 *
 * Unbind the item with the given tid from the current thread.
************************************************************************/


void
Tcb::Unbind( thread_t tid )
{
    assert(thr_curr);
	assert(thr_curr->i_own(tid));

	// Remove the Item from local and global waitset
	thr_all_waitset.Remove( tid );
	fd.waitset.Remove( tid );

	// ensure our "last" indicators are still valid
	if( thr_tab->IsFile( tid ) )
	{
		if( tid == fd.last_file )
		{
			fd.last_file = thr_tab->FindNext( tid, item_t_file );
			if( fd.last_file == tid )
			{
				fd.last_file = THR_TID_UNSPEC;
			}
		}
	}
	else if( thr_tab->IsSocket( tid ) )
	{
		if( tid == fd.last_socket )
		{
			fd.last_socket = thr_tab->FindNext( tid, item_t_socket );
			if( fd.last_socket == tid )
			{
				fd.last_socket = THR_TID_UNSPEC;
			}
		}
	}
}



#if defined(THREADLIB_DEBUG)
void
Tcb::WriteTo( ostream& s ) const
{
	s << "thread[" << id.tid << ',' << id.name << ':' << endl;

	// dump thread's wait set
	s << "  local waitset: " << fd.waitset << endl;

	// dump waiting status
    if (flags.blocked) {
        if (flags.polling) {
			s << "    polling for: (tid: " << *(msg.tidp)
				<< ", tag: " << *(msg.tagp)
				<< ')' << endl;
        }
        else {
			s << "    waiting for: (tid: " << *(msg.tidp)
				<< ", tag: " << *(msg.tagp)
				<< ", size: " << ((msg.nitemsp) ? *(msg.nitemsp) : 0)
				<< ')' << endl;
        }
    }

	// dump queue status
    qscan_t qs;
    ThrMessage* m;

    if (msg.mq) {
		s << "    mqueue:\n";
		qs = FIRST(msg.mq);
		while (!EOSCAN(msg.mq, qs)) {
			m = (ThrMessage*) GET(msg.mq, qs);
			s << m << endl;
			qs = NEXT(msg.mq, qs);
		}
    }
}
#endif // defined(THREADLIB_DEBUG)


bool
Tcb::i_own( thread_t tid ) const
{
	Item* pItem = thr_tab->GetItem(tid);

	return (pItem &&
			pItem->IsBindable() &&
			(thr_tab->GetBindable(tid)->tid == id.tid));
}





